from typing import Callable, List

from discord.ext.wizards.constants import ACTION


def action(*names: List[str]) -> Callable[[ACTION], ACTION]:
    def predicate(function: ACTION) -> ACTION:
        function.__action_triggers__ = names
        return function

    return predicate



from .action import action
from .step import Step, step
from .stopreason import StopReason
from .wizard import Wizard

__version__ = "0.2.2"

__all__ = [
    "Step",
    "step",
    "Wizard",
    "StopReason",
    "action",
]



import pathlib
from setuptools import setup

from discord.ext.wizards import __version__

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
print(type(README))

setup(
    name="discord-ext-wizards",
    author="Circuit",
    url="https://github.com/CircuitsBots/discord-ext-wizards",
    version=__version__,
    packages=["discord.ext.wizards"],
    license="MIT",
    description="A module for making setup wizards easily.",
    long_description=README,
    long_description_content_type="text/markdown",
    install_requires=["discord.py>=1.2.5"],
    python_requires=">=3.5.3",
)



import traceback
from typing import Any, Dict, List, Optional

from discord.ext.commands import Context

import discord
from discord.ext.wizards.constants import ACTION
from discord.ext.wizards.step import Step
from discord.ext.wizards.stopreason import StopReason


class Wizard:
    def __init__(self, cleanup_after: bool = True, timeout: float = 180.0):
        self._ctx: Optional[Context] = None
        self._running = True
        self._steps: List["Step"] = []
        self._actions: Dict[str, "ACTION"] = {}
        self._to_cleanup: List[int] = []

        self.timeout = timeout
        self.cleanup_after = cleanup_after
        self.stop_reason: Optional[StopReason] = None

        for attr_name in dir(self):
            attr = self.__getattribute__(attr_name)
            if isinstance(attr, Step):
                self._steps.append(attr)
            else:
                try:
                    names = attr.__action_triggers__
                except AttributeError:
                    pass
                else:
                    for n in names:
                        self._actions[n] = attr

        self._steps.sort(key=lambda s: s.index)

    async def start(self, ctx: Context) -> Dict[str, Any]:
        self._ctx = ctx
        await self._internal_loop()
        if self.cleanup_after:
            await self.cleanup()

    async def stop(self, reason: Optional[StopReason] = None):
        self.stop_reason = reason
        self._running = False

    async def cleanup(self):
        def check(m: discord.Message) -> bool:
            return m.id in self._to_cleanup

        try:
            await self._ctx.channel.purge(limit=200, check=check)
        except discord.Forbidden:
            for mid in self._to_cleanup:
                mobj = self._ctx.channel.get_partial_message(mid)
                try:
                    await mobj.delete()
                except (discord.NotFound, discord.Forbidden):
                    pass

    async def send(self, *args, **kwargs) -> discord.Message:
        m = await self._ctx.send(*args, **kwargs)
        self._to_cleanup.append(m.id)
        return m

    async def do_step(self, step: "Step") -> Any:
        return await step.do_step(self)

    async def on_step_error(self, step: "Step", err: Exception):
        traceback.print_exception(err.__class__, err, err.__traceback__)
        await self.stop(StopReason.ERROR)

    async def on_action_error(self, action: "ACTION", err: Exception):
        traceback.print_exception(err.__class__, err, err.__traceback__)
        await self.stop(StopReason.ERROR)

    async def _internal_loop(self):
        current_step = 0
        while self._running and current_step < len(self._steps):
            step = self._steps[current_step]
            current_step += 1
            if not step.call_internally:
                continue

            await step.do_step(self)
        self._running = False

    def _check_message(self, message: discord.Message) -> bool:
        return message.author.id == self._ctx.author.id



import asyncio
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from discord.ext.wizards.stopreason import StopReason
from discord.ext.wizards.constants import STEP

if TYPE_CHECKING:
    from discord.ext.wizards.wizard import Wizard


class Step:
    def __init__(
        self,
        index: int,
        action: STEP,
        name: str,
        call_internally: bool,
        description: Union[Callable[["Wizard"], str], str],
        timeout: Optional[float] = None,
    ):
        self.index = index
        self.action = action
        self.name = name
        self.description = description
        self.call_internally = call_internally
        self.timeout = timeout

    async def do_step(self, wizard: "Wizard") -> Any:
        try:
            return await self._do_step(wizard)
        except Exception as e:
            return await wizard.on_step_error(self, e)

    async def _do_step(self, wizard: "Wizard") -> Any:
        if isinstance(self.description, str):
            desc = self.description
        else:
            desc = self.description(wizard)
        await wizard.send(desc)
        try:
            message = await wizard._ctx.bot.wait_for(
                "message",
                check=wizard._check_message,
                timeout=self.timeout or wizard.timeout,
            )
        except asyncio.TimeoutError:
            return await wizard.stop(StopReason.TIMED_OUT)
        wizard._to_cleanup.append(message.id)
        if message.content in wizard._actions:
            action = wizard._actions[message.content]
            try:
                return await action(message)
            except Exception as e:
                return await wizard.on_action_error(action, e)
        return await self.action(wizard, message)


def step(
    description: Union[str, Callable[["Wizard"], str]],
    name: Optional[str] = None,
    call_internally: bool = True,
    position: int = None,
    timeout: Optional[float] = None,
) -> Callable[[STEP], Step]:
    def predicate(
        action: STEP,
    ) -> Step:
        return Step(
            position if position is not None else -1,
            action,
            name or action.__name__,
            call_internally,
            description,
            timeout,
        )

    return predicate
